// Copyright (c) 2015, Giliam de Carpentier
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the 
// documentation and/or other materials provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Class structure based on code from the Ogre3D SDK, Copyright (c) 2000-2012 Torus Knot Software Ltd.

#include "PrecompiledHeader.h"
#include "LensDistortionApplication.h"
#include "CustomCameraMan.h"
#include "LensDistortion.h"

static const Ogre::String gs_CompositorName("LensDistortion");

//-------------------------------------------------------------------------------------
LensDistortionApplication::LensDistortionApplication(void) 
{
}
//-------------------------------------------------------------------------------------
LensDistortionApplication::~LensDistortionApplication(void)
{
}

//-----------------------------------------------------------------------------------
void setVertexUniform(const Ogre::String& compositor, const Ogre::String& material, const Ogre::String& uniform, const Ogre::Vector4& value)
{
	Ogre::Pass* pass = (static_cast<Ogre::MaterialPtr>(Ogre::MaterialManager::getSingleton().getByName(material)))->getTechnique(0)->getPass(0);
	pass->getVertexProgramParameters()->setNamedConstant(uniform, value);
}

//-------------------------------------------------------------------------------------
void LensDistortionApplication::createScene(void)
{
	mSceneMgr->setShadowTechnique(Ogre::SHADOWTYPE_TEXTURE_MODULATIVE);
	mSceneMgr->setShadowTextureSize(2048);
	mSceneMgr->setShadowFarDistance(1000);
    
	Ogre::MovableObject::setDefaultVisibilityFlags(0x00000001);

	// ambient light
	mSceneMgr->setAmbientLight(Ogre::ColourValue(0.3f, 0.3f, 0.2f));

	// directional light
	Ogre::Light* l = mSceneMgr->createLight("Light2");
	Ogre::Vector3 dir(1,-2,0);
	dir.normalise();
	l->setType(Ogre::Light::LT_DIRECTIONAL);
	l->setDirection(dir);
	l->setDiffuseColour(1, 1, 0.8f);
	l->setSpecularColour(1, 1, 1);

	Ogre::Entity* entity;
	Ogre::SceneNode* node;
	
	// houses
	entity = mSceneMgr->createEntity( "1", "tudorhouse.mesh" );
	node = mSceneMgr->getRootSceneNode()->createChildSceneNode(Ogre::Vector3(437.5, 687.5, -600));
	node->attachObject( entity );
	node->scale(1.25f, 1.25f, 1.25f);

	entity = mSceneMgr->createEntity( "2", "tudorhouse.mesh" );
	node = mSceneMgr->getRootSceneNode()->createChildSceneNode(Ogre::Vector3(-437.5, 687.5, -600));
	node->attachObject( entity );
	node->scale(1.25f, 1.25f, 1.25f);
	
	// giant knot
	entity = mSceneMgr->createEntity( "3", "knot.mesh" );
	node = mSceneMgr->getRootSceneNode()->createChildSceneNode(Ogre::Vector3(125, 125, 375));
	node->setScale(67,67,67);
	node->attachObject( entity );
	entity->setMaterialName("Examples/MorningCubeMap");
	entity->setCastShadows(false);
		
	// row of shiny cubes of different sizes
	entity = mSceneMgr->createEntity( "boxm1", "cube.mesh" );
	node = mSceneMgr->getRootSceneNode()->createChildSceneNode(Ogre::Vector3(50, 150, 1000));
	node->setScale(0.125f, 0.125f, 0.125f);
	node->attachObject( entity );
	entity->setMaterialName("Examples/MorningCubeMap");
	
	entity = mSceneMgr->createEntity( "box0", "cube.mesh" );
	node = mSceneMgr->getRootSceneNode()->createChildSceneNode(Ogre::Vector3(100, 150, 1000));
	node->setScale(0.25f, 0.25f, 0.25f);
	node->attachObject( entity );
	entity->setMaterialName("Examples/MorningCubeMap");
			
	entity = mSceneMgr->createEntity( "box1", "cube.mesh" );
	node = mSceneMgr->getRootSceneNode()->createChildSceneNode(Ogre::Vector3(200, 150, 1000));
	node->setScale(0.5f, 0.5f, 0.5f);
	node->attachObject( entity );
	entity->setMaterialName("Examples/MorningCubeMap");
		
	entity = mSceneMgr->createEntity( "box2", "cube.mesh" );
	node = mSceneMgr->getRootSceneNode()->createChildSceneNode(Ogre::Vector3(400, 150, 1000));
	node->attachObject( entity );
	entity->setMaterialName("Examples/MorningCubeMap");
		
	entity = mSceneMgr->createEntity( "box3", "cube.mesh" );
	node = mSceneMgr->getRootSceneNode()->createChildSceneNode(Ogre::Vector3(800, 150, 1000));
	node->setScale(2, 2, 2);
	node->attachObject( entity );
	entity->setMaterialName("Examples/MorningCubeMap");

	entity = mSceneMgr->createEntity( "box4", "cube.mesh" );
	node = mSceneMgr->getRootSceneNode()->createChildSceneNode(Ogre::Vector3(1400, 150, 1000));
	node->setScale(4, 4, 4);
	node->attachObject( entity );
	entity->setMaterialName("Examples/MorningCubeMap");
	
	// vertical columns
	for (int x = -2; x <= 2; ++x)
	{
		for (int z = -2; z <= 2; ++z)
		{
			Ogre::StringUtil::StrStreamType str;
			str << "col" << x << "_" << z;
			Ogre::SceneNode* node = mSceneMgr->getRootSceneNode()->createChildSceneNode();
			entity = mSceneMgr->createEntity( str.str(), "column.mesh" );
			node->attachObject( entity );
			node->scale(0.5f,6.25f,0.5f);
			node->translate((0.866f*x-0.5f*z)*625, 656.25, (0.5f*x+0.866f*z)*625 + 2500);
		}
	}

	// horizontal columns
	for (int y = 0; y < 10; ++y)
	{
		Ogre::StringUtil::StrStreamType str;
		str << "col" << y;
		Ogre::SceneNode* node = mSceneMgr->getRootSceneNode()->createChildSceneNode();
		entity = mSceneMgr->createEntity( str.str(), "column.mesh" );
		node->attachObject( entity );
		node->lookAt(Ogre::Vector3(0.5f,0,-0.866f), Ogre::Node::TS_PARENT, Ogre::Vector3::UNIT_Y);
		node->scale(0.5f,6.25f,0.5f);
		node->translate(-0.866f*1250, 250*y, 3125);
	}

	// sky box
	mSceneMgr->setSkyBox(true, "Examples/MorningSkyBox");

	// ground plane
	{
		Ogre::Plane plane;
		plane.normal = Ogre::Vector3::UNIT_Y;
		plane.d = 0;
		Ogre::MeshManager::getSingleton().createPlane("Myplane",
			Ogre::ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME, plane,
			300000, 300000, 10, 10, true, 1, 2000, 2000, Ogre::Vector3::UNIT_Z);
		Ogre::Entity* pPlaneEnt = mSceneMgr->createEntity( "plane", "Myplane" );
		pPlaneEnt->setMaterialName("Examples/Rockwall");
		pPlaneEnt->setCastShadows(false);
		mSceneMgr->getRootSceneNode()->createChildSceneNode()->attachObject(pPlaneEnt);
	}

	// animated character
	{
		Ogre::Vector3 position(mCamera->getPosition() * Ogre::Vector3(1,0,1) + Ogre::Vector3(400.0f, 0.0f, 0.0f));

		entity = mSceneMgr->createEntity("JaiquaSneak", "jaiqua.mesh");
		entity->setMaterialName("jaiqua");

		node = mSceneMgr->getRootSceneNode()->createChildSceneNode();
		node->attachObject(entity);
		node->setPosition(position + Ogre::Vector3(0,10,0));
		node->setScale(11.0f, 11.0f, 11.0f);

		mCharacterAnimationState = entity->getAnimationState("Sneak");
		mCharacterAnimationState->setEnabled(true);
		mCharacterAnimationState->setLoop(true);
		entity->getSkeleton()->getRootBone()->setManuallyControlled(true);
		entity->getSkeleton()->getRootBone()->setPosition(Ogre::Vector3(8,0,5));
	}

	// character ring
	float angleStep = 3.141592f / 6;
	for (int i = -6; i < 6; ++i)
	{
		Ogre::StringUtil::StrStreamType str;
		str << "Jaiqua" << i;

		float angle = angleStep * (i + 0.5f);

		Ogre::Vector3 position(mCamera->getPosition() * Ogre::Vector3(1,0,1) + 180 * Ogre::Vector3(Ogre::Math::Sin(angle), 0, Ogre::Math::Cos(angle)));

		entity = mSceneMgr->createEntity(str.str(), "jaiqua.mesh");
		entity->setMaterialName("jaiqua");

		node = mSceneMgr->getRootSceneNode()->createChildSceneNode();
		node->attachObject(entity);
		node->setPosition(position + Ogre::Vector3(0,10,0));
		node->setOrientation(Ogre::Quaternion(Ogre::Radian(angle - 0.8f), Ogre::Vector3::UNIT_Y));
		node->setScale(11.0f, 11.0f, 11.0f);

		Ogre::AnimationState* as = entity->getAnimationState("Sneak");
		as->setEnabled(true);
		as->setLoop(true);
		entity->getSkeleton()->getRootBone()->setManuallyControlled(true);
		entity->getSkeleton()->getRootBone()->setPosition(Ogre::Vector3(8,0,5));
	}

	try 
	{
		Ogre::Viewport* viewport = mCamera->getViewport();
		Ogre::CompositorManager::getSingleton().addCompositor(viewport, gs_CompositorName);
		Ogre::CompositorManager::getSingleton().setCompositorEnabled(viewport, gs_CompositorName, true);

	} 
	catch (...) 
	{
		/// Warn user
		Ogre::LogManager::getSingleton().logMessage("Could not load compositor " + gs_CompositorName);
	}

	// build custom gui elements
	Ogre::StringVector paramNames;
	paramNames.push_back("FOVx min");
	paramNames.push_back("FOVx max");
	paramNames.push_back("FOVy min");
	paramNames.push_back("FOVy max");
	paramNames.push_back("FOV diag");
	paramNames.push_back("Visible");
	paramNames.push_back("Max stretch ratio");
	paramNames.push_back("Max scale");
	paramNames.push_back("Max curvature");
	paramNames.push_back("Min resolution");
	
	Ogre::FontManager::getSingleton().getByName("SdkTrays/Caption")->load();
	Ogre::FontManager::getSingleton().getByName("SdkTrays/Value")->load();

	mCameraPanel = mTrayMgr->createParamsPanel(OgreBites::TL_TOPLEFT, "CameraPanel", 200, paramNames);
	
	mFOVSlider = mTrayMgr->createThickSlider(OgreBites::TL_TOPLEFT, "FOVSlider", "FOVx at y = 1/3", 201, 50, 0, 160, 161);
	mFOVSlider->setValue(120.0f);
	
	mDistortionSlider = mTrayMgr->createThickSlider(OgreBites::TL_TOPLEFT, "DistortionSlider", "Distortion", 201, 100, 0, 1.0f, 101);
	mDistortionSlider->setValue(0.5f);

	mAutoDirectionCheckbox = mTrayMgr->createCheckBox(OgreBites::TL_TOPLEFT, "AutoDirectionCheckbox", "Auto direction", 201);
	mAutoDirectionCheckbox->setChecked(true);
		
	mCylindericalRatioSlider = mTrayMgr->createThickSlider(OgreBites::TL_TOPLEFT, "CylindericalSlider", "Cylinderical ratio", 201, 50, 0.25f, 4.0f, 76);
	mCylindericalRatioSlider->setValue(2.0f);

	mSkipPostEffectCheckbox = mTrayMgr->createCheckBox(OgreBites::TL_TOPLEFT, "SkipPostEffectCheckbox", "Skip PostFX", 201);
	mSkipPostEffectCheckbox->setChecked(false);
}

//-----------------------------------------------------------------------------------
bool LensDistortionApplication::keyPressed( const OIS::KeyEvent &arg )
{
	bool result = BaseApplication::keyPressed(arg);
	return result;
}

//-----------------------------------------------------------------------------------
bool LensDistortionApplication::mousePressed( const OIS::MouseEvent &arg, OIS::MouseButtonID id )
{
	bool result = BaseApplication::mousePressed(arg, id);
	return result;
}

//-----------------------------------------------------------------------------------
bool LensDistortionApplication::frameRenderingQueued(const Ogre::FrameEvent& evt)
{
	mCharacterAnimationState->addTime(evt.timeSinceLastFrame);

	Ogre::Vector3 position(mCameraMan->getPivotPosition());
	mCameraMan->setPivotPosition(position);

	float cylindricalRatio;
	if (mAutoDirectionCheckbox->isChecked())
	{
		// In this mode, the cylindrical ratio set with the slider is scaled based on the current camera's pitch. That is, the set cylindrical ratio is
		// used fully when purely looking at the horizontal plane, but it's scaled towards 1.0 when the camera is aimed more and more vertically (either up or down).
		float mCameraPitchCos = Ogre::Math::Sqrt(mCamera->getDirection().x * mCamera->getDirection().x + mCamera->getDirection().z * mCamera->getDirection().z);
		cylindricalRatio = 1.0f + (mCylindericalRatioSlider->getValue() - 1.0f) * mCameraPitchCos * mCameraPitchCos;
	}
	else
	{
		cylindricalRatio = mCylindericalRatioSlider->getValue();
	}

	Ogre::Radian fovDegrees(Ogre::Degree(mFOVSlider->getValue()));

	LensDistortion::AspectRatio aspectRatio(mCamera->getAspectRatio());
	LensDistortion::AspectRatio cylindricalAspectRatio(mCamera->getAspectRatio() * cylindricalRatio);

	float tan = LensDistortion::fovDegreesToFOVTan(mFOVSlider->getValue());
	float diagonalFOVTan;

	{
		// Below, pointOnNonNormalizedRectangleToDiagonalFOVTan() is used to calculate the diagonal fov required to get exactly the horizontal fov (as 
		// specified by the slider) at a given horizontal line on the screen for which y=1/3 (where y = 0 is mid-screen and y = 1 is at the top of the screen). 
		// In the article, this procedure is called 'pinning the horizontal fov'.
		float yAtWhichFOVIsSpecified = 0.333f;
		float horizontalFOVTan = LensDistortion::fovDegreesToFOVTan(mFOVSlider->getValue());
		diagonalFOVTan = LensDistortion::pointOnNonNormalizedRectangleToDiagonalFOVTan(mDistortionSlider->getValue(), cylindricalAspectRatio, 
			horizontalFOVTan, yAtWhichFOVIsSpecified * cylindricalAspectRatio.horizontalToVerticalTan(horizontalFOVTan));

		if (diagonalFOVTan < 0)
		{
			diagonalFOVTan = 100.0f;
		}
	}

	LensDistortion::NormalizedDistortion distortion(aspectRatio, cylindricalRatio, mDistortionSlider->getValue(), diagonalFOVTan);
	updateCameraFOV(distortion);
	updateStatistics(distortion);
	updateLensDistortionShaderParameters(distortion);

	return BaseApplication::frameRenderingQueued(evt);
}

void LensDistortionApplication::updateCameraFOV(const LensDistortion::NormalizedDistortion& distortion)
{
	Ogre::Radian fovYRadians(LensDistortion::fovTanToFOVRadians(distortion.getAspectRatio().diagonalToVerticalTan(distortion.getDiagonalFOVTan())));
	mCamera->setFOVy(fovYRadians);
}

void LensDistortionApplication::updateStatistics(const LensDistortion::NormalizedDistortion& distortion)
{
	LensDistortion::AspectRatio aspectRatio(distortion.getAspectRatio());
	LensDistortion::REAL diagonalFOVTan = distortion.getDiagonalFOVTan();
	LensDistortion::REAL fovDiagonalMaxHorizontal = LensDistortion::fovTanToFOVDegrees(aspectRatio.diagonalToHorizontalTan(diagonalFOVTan));
	LensDistortion::REAL fovDiagonalMaxVertical = LensDistortion::fovTanToFOVDegrees(aspectRatio.diagonalToVerticalTan(diagonalFOVTan));
	LensDistortion::REAL fovDiagonalDegrees = LensDistortion::fovTanToFOVDegrees(diagonalFOVTan);

	LensDistortion::REAL fovDiagonalMinHorizontal;
	{
		LensDistortion::REAL x = 1, y = 0;
		LensDistortion::undistort(distortion, x, y);
		LensDistortion::unnormalizeNonDistorted(distortion, x, y);
		fovDiagonalMinHorizontal = LensDistortion::fovTanToFOVDegrees(x);
	}

	LensDistortion::REAL fovDiagonalMinVertical;
	{
		LensDistortion::REAL x = 0, y = 1;
		LensDistortion::undistort(distortion, x, y);
		LensDistortion::unnormalizeNonDistorted(distortion, x, y);
		fovDiagonalMinVertical = LensDistortion::fovTanToFOVDegrees(y);
	}

	float coveragedArea = LensDistortion::Properties::coveredArea(distortion);
	float maxStretch = LensDistortion::Properties::maxStretch(distortion);
	float maxScaling = LensDistortion::Properties::maxScaling(distortion);
	float maxCurvatureRelativeToDiameter = 0.5f * LensDistortion::Properties::maxCurvature(distortion);
	float minResolution = 1.0f / distortion.getZoomCoeff();

	mCameraPanel->setParamValue(0, Ogre::StringConverter::toString((float)fovDiagonalMinHorizontal));
	mCameraPanel->setParamValue(1, Ogre::StringConverter::toString((float)fovDiagonalMaxHorizontal));
	mCameraPanel->setParamValue(2, Ogre::StringConverter::toString((float)fovDiagonalMinVertical));
	mCameraPanel->setParamValue(3, Ogre::StringConverter::toString((float)fovDiagonalMaxVertical));
	mCameraPanel->setParamValue(4, Ogre::StringConverter::toString((float)fovDiagonalDegrees));
	mCameraPanel->setParamValue(5, Ogre::StringConverter::toString((int)(100.0f * coveragedArea)) + "%");
	mCameraPanel->setParamValue(6, Ogre::StringConverter::toString((int)(100.0f * maxStretch)) + "%");
	mCameraPanel->setParamValue(7, Ogre::StringConverter::toString((int)(100.0f * maxScaling)) + "%");
	mCameraPanel->setParamValue(8, Ogre::StringConverter::toString((int)(100.0f * maxCurvatureRelativeToDiameter)) + "%");
	mCameraPanel->setParamValue(9, Ogre::StringConverter::toString((int)(100.0f * minResolution)) + "%");
}

void LensDistortionApplication::updateLensDistortionShaderParameters(const LensDistortion::NormalizedDistortion& distortion)
{
	Ogre::Vector4 uvScales;	

	uvScales.x = 2.0f * distortion.getCylindricalAspectRatio().getRatio() * Ogre::Math::Sqrt(distortion.getNormCoeff());
	uvScales.y = 2.0f * Ogre::Math::Sqrt(distortion.getNormCoeff());
	uvScales.z = distortion.getZoomCoeff();
	uvScales.w = 0.5f * distortion.getZoomCoeff() - 0.5f;

	setVertexUniform(gs_CompositorName, "Ogre/Compositor/LensDistortionPass", "uvScales", uvScales);
	Ogre::Viewport* viewport = mCamera->getViewport();
	Ogre::CompositorManager::getSingleton().setCompositorEnabled(viewport, gs_CompositorName, false);

	if (!mSkipPostEffectCheckbox->isChecked())
	{
		Ogre::CompositorManager::getSingleton().setCompositorEnabled(viewport, gs_CompositorName, true);
	}
}

#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
#define WIN32_LEAN_AND_MEAN
#include "windows.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
    INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR strCmdLine, INT )
#else
    int main(int argc, char *argv[])
#endif
    {
        // Create application object
        LensDistortionApplication app;

        try 
		{
            app.go();
        } 
		catch( Ogre::Exception& e ) 
		{
#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
            MessageBox( NULL, e.getFullDescription().c_str(), "An exception has occured!", MB_OK | MB_ICONERROR | MB_TASKMODAL);
#else
            std::cerr << "An exception has occured: " <<
                e.getFullDescription().c_str() << std::endl;
#endif
        }

        return 0;
    }

#ifdef __cplusplus
}
#endif
